import json
from os import devnull, environ
from subprocess import Popen, PIPE

ARR_START = "[".encode()
OBJ_START = "{".encode()


def az_login():
    proc = Popen(
        ["az", "login"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    if proc.returncode != 0:
        return []
    return json.loads(output[output.find(ARR_START) :])


def az_logout():
    proc = Popen(
        ["az", "logout"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    return proc.returncode == 0


def az_account_show():
    proc = Popen(
        ["az", "account", "show"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, _err = proc.communicate()
    if proc.returncode != 0:
        return {}
    return json.loads(output[output.find(OBJ_START) :])


def az_account_list():
    proc = Popen(
        ["az", "account", "list"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    if proc.returncode != 0:
        return []
    return json.loads(output[output.find(ARR_START) :])


def az_find_subscription_id(name, account_list):
    for subscription in account_list:
        if subscription["name"] == name or subscription["id"] == name:
            return subscription["id"]
    return None


def az_set_subscription(subscription_id):
    proc = Popen(
        ["az", "account", "set", "--subscription", subscription_id],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    return proc.returncode == 0


def az_select_subscription(name):
    curr_account = az_account_show()
    subscription_id = None
    if "name" in curr_account and "id" in curr_account:
        if curr_account["name"] == name or curr_account["id"] == name:
            return curr_account["id"]
        else:
            subscription_id = az_find_subscription_id(name, az_account_list())
    if not subscription_id:
        az_logout()
        subscription_id = az_find_subscription_id(name, az_login())

    if subscription_id:
        az_set_subscription(subscription_id)
    return subscription_id


def ensure_group(location, group_name):
    proc = Popen(
        ["az", "group", "list", "--query", "[?name=='" + group_name + "']"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    groups = json.loads(output[output.find(ARR_START) :])
    if len(groups) == 0:
        proc = Popen(
            ["az", "group", "create", "--name", group_name, "--location", location],
            stdout=PIPE,
            stderr=PIPE,
        )
        output, err = proc.communicate()
        if proc.returncode == 0:
            groups = [json.loads(output[output.find(OBJ_START) :])]
        else:
            groups = [{}]
    return groups[0]


def ensure_management_group(group_name):
    proc = Popen(
        [
            "az",
            "account",
            "management-group",
            "list",
            "--query",
            "[?name=='" + group_name + "']",
        ],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    groups = json.loads(output[output.find(ARR_START) :])
    if len(groups) == 0:
        proc = Popen(
            ["az", "account", "management-group", "create", "--name", group_name],
            stdout=PIPE,
            stderr=PIPE,
        )
        output, err = proc.communicate()
        if proc.returncode == 0:
            groups = [json.loads(output[output.find(OBJ_START) :])]
        else:
            groups = [{}]


def delete_group(group_name):
    proc = Popen(
        ["az", "group", "delete", "-y", "--name", group_name],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    return proc.returncode == 0



from locale import getpreferredencoding
from os import linesep, environ, devnull
from sys import argv
from n_utils.profile_util import enable_profile
from subprocess import Popen, PIPE

SYS_ENCODING = getpreferredencoding()


def _var_name(suffix):
    return "ndt.profile." + suffix


def _branch_var_name(current_branch, suffix):
    return "ndt.profile." + current_branch + "." + suffix


def _check_enable_profile(vars, current_branch, profile_type):
    if _branch_var_name(current_branch, profile_type) in vars:
        enable_profile(
            profile_type, vars[_branch_var_name(current_branch, profile_type)]
        )
    elif _var_name(profile_type) in vars:
        enable_profile(profile_type, vars[_var_name(profile_type)])


def _check_enable_virtualenv(vars, current_branch):
    pyenv = None
    if "ndt.virtualenv." + current_branch in vars:
        pyenv = vars["ndt.virtualenv." + current_branch]
    elif "ndt.virtualenv" in vars:
        pyenv = vars["ndt.virtualenv"]
    if pyenv:
        if not ("PYENV_VERSION" in environ and environ["PYENV_VERSION"] == pyenv):
            print("export PYENV_VIRTUALENV_DISABLE_PROMPT=1;")
            print("pyenv activate '" + pyenv + "'")
    else:
        _check_disable_virtualenv()


def _check_disable_virtualenv():
    if "PYENV_VERSION" in environ and environ["PYENV_VERSION"].startswith("ndt-"):
        print("pyenv deactivate;")
        print("unset PYENV_VIRTUALENV_DISABLE_PROMPT")


def load_project_env():
    """Print parameters set by git config variables to setup project environment with region and aws credentials"""
    proc = Popen(
        ["git", "config", "--list", "--local"], stdout=PIPE, stderr=open(devnull, "w")
    )
    proc2 = Popen(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        stdout=PIPE,
        stderr=open(devnull, "w"),
    )
    stdout, _ = proc2.communicate()
    current_branch = "master"
    if proc2.returncode == 0:
        current_branch = stdout.decode(SYS_ENCODING).strip()
    out, _ = proc.communicate()
    if proc.returncode:
        _check_disable_virtualenv()
        return
    vars = {}
    for line in out.decode(SYS_ENCODING).split("\n"):
        if line:
            next = line.split("=", 1)
            vars[next[0]] = next[1]
    do_print = False
    ret = ""

    _check_enable_profile(vars, current_branch, "azure")
    _check_enable_profile(vars, current_branch, "adfs")
    _check_enable_profile(vars, current_branch, "lastpass")
    _check_enable_profile(vars, current_branch, "iam")
    _check_enable_profile(vars, current_branch, "ndt")
    _check_enable_profile(vars, current_branch, "azure-subscription")
    _check_enable_virtualenv(vars, current_branch)

    if "ndt.source." + current_branch + ".env" in vars:
        do_print = True
        ret = ret + ". " + vars["ndt.source." + current_branch + ".env"] + linesep
    elif "ndt.source.env" in vars:
        do_print = True
        ret = ret + ". " + vars["ndt.source.env"] + linesep

    if "ndt.aws." + current_branch + ".profile" in vars:
        do_print = True
        ret = (
            ret
            + "export AWS_PROFILE="
            + vars["ndt.aws." + current_branch + ".profile"]
            + " AWS_DEFAULT_PROFILE="
            + vars["ndt.aws." + current_branch + ".profile"]
            + linesep
        )
    elif "ndt.aws.profile" in vars:
        do_print = True
        ret = (
            ret
            + "export AWS_PROFILE="
            + vars["ndt.aws.profile"]
            + " AWS_DEFAULT_PROFILE="
            + vars["ndt.aws.profile"]
            + linesep
        )

    if "ndt.aws." + current_branch + ".region" in vars:
        do_print = True
        ret = (
            ret
            + "export AWS_REGION="
            + vars["ndt.aws." + current_branch + ".region"]
            + " AWS_DEFAULT_REGION="
            + vars["ndt.aws. " + current_branch + ".region"]
            + linesep
        )
    elif "ndt.aws.region" in vars:
        do_print = True
        ret = (
            ret
            + "export AWS_REGION="
            + vars["ndt.aws.region"]
            + " AWS_DEFAULT_REGION="
            + vars["ndt.aws.region"]
            + linesep
        )

    if do_print:
        print(ret.strip())


def ndt_register_complete():
    """Print out shell function and command to register ndt command completion"""
    print(
        """_ndt_complete() {
    local IFS=$'\\013'
    local COMP_CUR="${COMP_WORDS[COMP_CWORD]}"
    local COMP_PREV="${COMP_WORDS[COMP_CWORD-1]}"
    local SUPPRESS_SPACE=0
    if compopt +o nospace 2> /dev/null; then
        SUPPRESS_SPACE=1
    fi
    COMPREPLY=( $(IFS="$IFS" \\
                  COMP_LINE="$COMP_LINE" \\
                  COMP_POINT="$COMP_POINT" \\
                  COMP_TYPE="$COMP_TYPE" \\
                  COMP_CUR="$COMP_CUR" \\
                  COMP_PREV="$COMP_PREV" \\
                  COMP_CWORD=$COMP_CWORD \\
                  _ARGCOMPLETE_COMP_WORDBREAKS="$COMP_WORDBREAKS" \\
                  _ARGCOMPLETE=1 \\
                  _ARGCOMPLETE_SUPPRESS_SPACE=$SUPPRESS_SPACE \\
                  "$1" 8>&1 9>&2 1>/dev/null 2>/dev/null) )
    if [[ $? != 0 ]]; then
        unset COMPREPLY
    elif [[ $SUPPRESS_SPACE == 1 ]] && [[ "$COMPREPLY" =~ [=/:]$ ]]; then
        compopt -o nospace
    fi
}
complete -o nospace -F _ndt_complete "ndt"
"""
    )
    if len(argv) > 1 and "--project-env" in argv:
        print(
            """_projectenv_hook() {
  local previous_exit_status=$?;
  eval "$(nameless-dt-load-project-env)";
  return $previous_exit_status;
};
if ! [[ "$PROMPT_COMMAND" =~ _projectenv_hook ]]; then
  PROMPT_COMMAND="_projectenv_hook;$PROMPT_COMMAND";
fi
"""
        )
    if len(argv) > 1 and "--nep-function" in argv:
        print(
            """_nep_complete() {
    COMPREPLY=( $(ndt print-aws-profiles "${COMP_WORDS[COMP_CWORD]}" ) )
}

nep() {
    if [ "$1" = "-h" -o "$1" = "--help" ]; then
        ndt enable-profile -h
        return
    fi
    eval "$(ndt enable-profile $@)"
}

complete -F _nep_complete nep
"""
        )



import argparse
import os
import re
import json
from datetime import datetime
from os import R_OK, access
from os.path import exists, expanduser, isfile, join

import argcomplete
from argcomplete.completers import ChoicesCompleter
from dateutil.parser import parse
from dateutil.tz import tzutc

from n_utils import _to_str
from n_utils.bw_util import get_bwentry
from n_utils.lp_util import get_lpentry
from n_utils.az_util import az_select_subscription
from n_utils.mfa_utils import mfa_generate_code


def ConfigParser():
    import configparser

    return configparser.ConfigParser()


def read_expiring_profiles():
    ret = []
    home = expanduser("~")
    credentials = join(home, ".aws", "credentials")
    if exists(credentials):
        parser = ConfigParser()
        with open(credentials) as credfile:
            parser.read_file(credfile)
            for profile in parser.sections():
                if parser.has_option(
                    profile, "aws_session_expiration"
                ) or parser.has_option(profile, "aws_expiration"):
                    ret.append(profile)
    return ret


def read_profiles(prefix=None):
    ret = []
    home = expanduser("~")
    credentials = join(home, ".aws", "credentials")
    if isfile(credentials) and access(credentials, R_OK):
        parser = ConfigParser()
        with open(credentials) as credfile:
            parser.read_file(credfile)
            for profile in parser.sections():
                if profile.startswith(prefix):
                    ret.append(profile)
    config = join(home, ".aws", "config")
    if isfile(config) and access(config, R_OK):
        parser = ConfigParser()
        with open(config) as configfile:
            parser.read_file(configfile)
            for profile in parser.sections():
                if (
                    profile.startswith("profile ")
                    and profile[8:] not in ret
                    and profile[8:].startswith(prefix)
                ):
                    ret.append(profile[8:])
    return ret


def get_profile(profile, include_creds=True):
    home = expanduser("~")
    from collections import OrderedDict

    ret = OrderedDict()
    config = join(home, ".aws", "config")
    profile_section = "profile " + profile
    if isfile(config) and access(config, R_OK):
        parser = ConfigParser()
        with open(config) as configfile:
            parser.read_file(configfile)
            if profile_section in parser.sections():
                for option in parser.options(profile_section):
                    ret[option] = parser.get(profile_section, option)
    if include_creds:
        credentials = join(home, ".aws", "credentials")
        if isfile(credentials) and access(credentials, R_OK):
            parser = ConfigParser()
            with open(credentials) as credfile:
                parser.read_file(credfile)
                if profile in parser.sections():
                    for option in parser.options(profile):
                        ret[option] = parser.get(profile, option)
    return ret


def read_profile_expiry(profile):
    home = expanduser("~")
    credentials = join(home, ".aws", "credentials")
    if exists(credentials):
        parser = ConfigParser()
        with open(credentials) as credfile:
            parser.read_file(credfile)
            if parser.has_option(profile, "aws_expiration"):
                return parser.get(profile, "aws_expiration")
            elif parser.has_option(profile, "aws_session_expiration"):
                return parser.get(profile, "aws_session_expiration")
    return "1970-01-01T00:00:00.000Z"


def read_profile_expiry_epoc(profile):
    return _epoc_secs(parse(read_profile_expiry(profile)).replace(tzinfo=tzutc()))


def print_aws_profiles():
    """Prints profile names from credentials file (~/.aws/credentials) and the conf file (~/.aws/conf) for autocomplete tools"""
    parser = argparse.ArgumentParser(description=print_aws_profiles.__doc__)
    if "_ARGCOMPLETE" in os.environ:
        parser.add_argument(
            "prefix", help="Prefix of profiles to print", default="", nargs="?"
        ).completer = ChoicesCompleter(read_profiles())
        argcomplete.autocomplete(parser)
    else:
        parser.add_argument(
            "prefix", help="Prefix of profiles to print", default="", nargs="?"
        )
    args = parser.parse_args()
    print(" ".join(read_profiles(prefix=args.prefix)))


def profile_to_env():
    """Prints profile parameters from credentials file (~/.aws/credentials) as eval-able environment variables"""
    parser = argparse.ArgumentParser(description=profile_to_env.__doc__)
    parser.add_argument(
        "-t",
        "--target-role",
        action="store_true",
        help="Output also azure_default_role_arn",
    )
    parser.add_argument(
        "-r",
        "--role-arn",
        help="Output also the role given here as the target role for the profile",
    )
    if "_ARGCOMPLETE" in os.environ:
        parser.add_argument(
            "profile", help="The profile to read profile info from"
        ).completer = ChoicesCompleter(read_profiles())
        argcomplete.autocomplete(parser)
    else:
        parser.add_argument("profile", help="The profile to read profile info from")
    args = parser.parse_args()
    safe_profile = re.sub("[^A-Z0-9]", "_", args.profile.upper())
    params = []
    role_param = "AWS_TARGET_ROLE_ARN_" + safe_profile
    if args.target_role:
        profile_entry = "profile " + args.profile
        home = expanduser("~")
        config = join(home, ".aws", "config")
        if exists(config):
            parser = ConfigParser()
            with open(config) as configfile:
                parser.read_file(configfile)
                if profile_entry in parser.sections() and parser.has_option(
                    profile_entry, "azure_default_role_arn"
                ):
                    params.append(role_param)
                    print(
                        role_param
                        + '="'
                        + parser.get(profile_entry, "azure_default_role_arn")
                        + '";'
                    )
                if profile_entry in parser.sections() and parser.has_option(
                    profile_entry, "adfs_role_arn"
                ):
                    params.append(role_param)
                    print(
                        role_param
                        + '="'
                        + parser.get(profile_entry, "adfs_role_arn")
                        + '";'
                    )
                if profile_entry in parser.sections() and parser.has_option(
                    profile_entry, "lastpass_role_arn"
                ):
                    params.append(role_param)
                    print(
                        role_param
                        + '="'
                        + parser.get(profile_entry, "lastpass_role_arn")
                        + '";'
                    )
    if args.role_arn:
        params.append(role_param)
        print(role_param + '="' + args.role_arn + '";')
    print_profile(args.profile, params)


def print_profile(profile_name, params):
    safe_profile = re.sub("[^A-Z0-9]", "_", profile_name.upper())
    profile = get_profile(profile_name)
    for key, value in list(profile.items()):
        upper_param = key.upper()
        if key == "aws_session_expiration" or key == "aws_expiration":
            d = parse(value)
            print(
                "AWS_SESSION_EXPIRATION_EPOC_"
                + safe_profile
                + '="'
                + _to_str(_epoc_secs(d))
                + '";'
            )
            params.append("AWS_SESSION_EXPIRATION_EPOC_" + safe_profile)
        params.append(upper_param)
        if value.startswith('"'):
            value = value[1:-1]
        print(upper_param + '="' + value + '";')
    print("export " + " ".join(params) + ";")


def profile_expiry_to_env():
    """Prints profile expiry from credentials file (~/.aws/credentials) as eval-able environment variables"""
    parser = argparse.ArgumentParser(description=profile_expiry_to_env.__doc__)
    if "_ARGCOMPLETE" in os.environ:
        parser.add_argument(
            "profile", help="The profile to read expiry info from"
        ).completer = ChoicesCompleter(read_expiring_profiles())
        argcomplete.autocomplete(parser)
    else:
        parser.add_argument("profile", help="The profile to read expiry info from")
    args = parser.parse_args()
    print_profile_expiry(args.profile)


def print_profile_expiry(profile):
    safe_profile = re.sub("[^A-Z0-9]", "_", profile.upper())
    expiry = read_profile_expiry(profile)
    epoc = _epoc_secs(parse(expiry).replace(tzinfo=tzutc()))
    print("AWS_SESSION_EXPIRATION_EPOC_" + safe_profile + "=" + _to_str(epoc))
    print("AWS_SESSION_EXPIRATION_" + safe_profile + "=" + expiry)
    print(
        "export AWS_SESSION_EXPIRATION_"
        + safe_profile
        + " AWS_SESSION_EXPIRATION_EPOC_"
        + safe_profile
        + ";"
    )


def cli_read_profile_expiry():
    """Read expiry field from credentials file, which is there if the login happened
    with aws-azure-login or another tool that implements the same logic (e.g. https://github.com/NitorCreations/adfs-aws-login)."""
    parser = argparse.ArgumentParser(description=cli_read_profile_expiry.__doc__)
    parser.add_argument(
        "profile", help="The profile to read expiry info from"
    ).completer = ChoicesCompleter(read_expiring_profiles())
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    print(read_profile_expiry(args.profile))


def update_profile(profile, creds):
    home = expanduser("~")
    credentials = join(home, ".aws", "credentials")
    if exists(credentials):
        parser = ConfigParser()
        with open(credentials, "r") as credfile:
            parser.read_file(credfile)
            if profile not in parser.sections():
                parser.add_section(profile)
            parser.set(profile, "aws_access_key_id", creds["AccessKeyId"])
            parser.set(profile, "aws_secret_access_key", creds["SecretAccessKey"])
            parser.set(profile, "aws_session_token", creds["SessionToken"])
            parser.set(
                profile,
                "aws_session_expiration",
                creds["Expiration"].strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            )
            parser.set(
                profile,
                "aws_expiration",
                creds["Expiration"].strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            )
    with open(credentials, "w") as credfile:
        parser.write(credfile)


def cli_profiles_to_json():
    """Prints aws config file contents as json for further parsing and use in other tools"""
    parser = argparse.ArgumentParser(description=cli_profiles_to_json.__doc__)
    argcomplete.autocomplete(parser)
    _ = parser.parse_args()
    profiles_to_json()


def profiles_to_json():
    from collections import OrderedDict

    home = expanduser("~")
    config = join(home, ".aws", "config")
    ret = OrderedDict()
    if exists(config):
        parser = ConfigParser()
        with open(config, "r") as conffile:
            parser.read_file(conffile)
            for section in parser.sections():
                if section.startswith("profile "):
                    section_data = OrderedDict()
                    ret[section[8:]] = section_data
                    for option in parser.options(section):
                        section_data[option] = parser.get(section, option)
        print(json.dumps(ret, indent=2))


def update_profile_conf(profile, creds):
    home = expanduser("~")
    config = join(home, ".aws", "config")
    profile_section = "profile " + profile
    if exists(config):
        parser = ConfigParser()
        with open(config, "r") as conffile:
            parser.read_file(conffile)
            if profile_section not in parser.sections():
                parser.add_section(profile_section)
            for key in creds:
                parser.set(profile_section, key, creds[key])
    with open(config, "w") as conffile:
        parser.write(conffile)
    return


def store_bw_profile(bw_entry_name):
    bw_entry = get_bwentry(bw_entry_name)
    if (
        bw_entry
        and "aws_access_key_id" in bw_entry.fields
        and "aws_secret_access_key" in bw_entry.fields
        and "profile_name" in bw_entry.fields
    ):
        profile = bw_entry.fields["profile_name"]
        home = expanduser("~")
        credentials = join(home, ".aws", "credentials")
        if exists(credentials):
            parser = ConfigParser()
            with open(credentials, "r") as credfile:
                parser.read_file(credfile)
                if profile not in parser.sections():
                    parser.add_section(profile)
                parser.set(
                    profile, "aws_access_key_id", bw_entry.fields["aws_access_key_id"]
                )
                parser.set(
                    profile,
                    "aws_secret_access_key",
                    bw_entry.fields["aws_secret_access_key"],
                )
        with open(credentials, "w") as credfile:
            parser.write(credfile)
        return profile
    return None


def cli_store_bw_profile():
    """Fetches a Bitwarde entry and if it contains a definition of a aws credentials, stores it in aws cli
    configuration. Namely the entry needs to define the extra fields aws_access_key_id, aws_secret_access_key
    and profile_name"""
    parser = argparse.ArgumentParser(description=cli_store_bw_profile.__doc__)
    parser.add_argument("entryname", help="The name of the bitwarden entry to get")
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    profile = store_bw_profile(args.entryname)
    if profile:
        print("Saved profile " + profile)
    else:
        print("Entry " + args.entryname + " not found to contain an aws cli profile")


def cli_enable_profile():
    """Enable a configured profile. Simple IAM user, AzureAD, ADFS and ndt assume-role profiles are supported"""
    parser = argparse.ArgumentParser(description=cli_enable_profile.__doc__)
    type_select = parser.add_mutually_exclusive_group(required=False)
    type_select.add_argument(
        "-i", "--iam", action="store_true", help="IAM user type profile"
    )
    type_select.add_argument(
        "-a", "--azure", action="store_true", help="Azure login type profile"
    )
    type_select.add_argument(
        "-f", "--adfs", action="store_true", help="ADFS login type profile"
    )
    type_select.add_argument(
        "-l", "--lastpass", action="store_true", help="Lastpass login type profile"
    )
    type_select.add_argument(
        "-n", "--ndt", action="store_true", help="NDT assume role type profile"
    )
    type_select.add_argument(
        "-s",
        "--azure-subscription",
        action="store_true",
        help="Microsoft Azure subscription",
    )
    if "_ARGCOMPLETE" in os.environ:
        parser.add_argument(
            "profile", help="The profile to enable"
        ).completer = ChoicesCompleter(read_profiles())
        argcomplete.autocomplete(parser)
    else:
        parser.add_argument("profile", help="The profile to enable")
    args = parser.parse_args()
    if args.iam:
        profile_type = "iam"
    elif args.azure:
        profile_type = "azure"
    elif args.adfs:
        profile_type = "adfs"
    elif args.lastpass:
        profile_type = "lastpass"
    elif args.ndt:
        profile_type = "ndt"
    elif args.azure_subscription:
        profile_type = "azure-subscription"
    else:
        profile_type = resolve_profile_type(args.profile)
    enable_profile(profile_type, args.profile)


def resolve_profile_type(profile_name):
    profile = get_profile(profile_name)
    if "azure_tenant_id" in profile:
        profile_type = "azure"
    elif "ndt_role_arn" in profile:
        profile_type = "ndt"
    elif "adfs_login_url" in profile:
        profile_type = "adfs"
    elif "lastpass_saml_id" in profile:
        profile_type = "lastpass"
    else:
        profile_type = "iam"
    return profile_type


def enable_profile(profile_type, profile):
    orig_profile = profile
    profile = re.sub("[^a-zA-Z0-9_-]", "_", profile)
    safe_profile = re.sub("[^A-Z0-9]", "_", profile.upper())
    now = _epoc_secs(datetime.now(tzutc()))
    expiry = now - 1000
    if profile_type == "iam":
        _print_profile_switch(profile)
    elif (
        profile_type == "azure" or profile_type == "adfs" or profile_type == "lastpass"
    ):
        _print_profile_switch(profile)
        if "AWS_SESSION_EXPIRATION_EPOC_" + safe_profile in os.environ:
            expiry = int(os.environ["AWS_SESSION_EXPIRATION_EPOC_" + safe_profile])
        if expiry < now:
            expiry = read_profile_expiry_epoc(profile)
        if expiry < now:
            bw_prefix = ""
            bw_entry = None
            lp_entry = None
            profile_data = get_profile(profile)
            if "bitwarden_entry" in profile_data and profile_data["bitwarden_entry"]:
                bw_entry = get_bwentry(profile_data["bitwarden_entry"])
            if "lastpass_entry" in profile_data and profile_data["lastpass_entry"]:
                lp_entry = get_lpentry(profile_data["lastpass_entry"])
            if "AWS_SESSION_EXPIRATION_EPOC_" + safe_profile in os.environ:
                print("unset AWS_SESSION_EXPIRATION_EPOC_" + safe_profile + ";")
            if profile_type == "azure":
                gui_mode = ""
                if "azure_login_mode" in profile_data:
                    gui_mode = " --mode=" + profile_data["azure_login_mode"]
                if bw_entry:
                    bw_prefix = "AZURE_DEFAULT_PASSWORD='" + bw_entry.password + "' "
                elif lp_entry:
                    bw_prefix = "AZURE_DEFAULT_PASSWORD='" + lp_entry.password + "' "
                print(
                    bw_prefix
                    + "aws-azure-login --profile "
                    + profile
                    + gui_mode
                    + " --no-prompt --no-sandbox"
                )
            elif profile_type == "adfs":
                if bw_entry:
                    bw_prefix = "ADFS_DEFAULT_PASSWORD='" + bw_entry.password + "' "
                elif lp_entry:
                    bw_prefix = "ADFS_DEFAULT_PASSWORD='" + lp_entry.password + "' "
                print(
                    bw_prefix + "adfs-aws-login --profile " + profile + " --no-prompt"
                )
            elif profile_type == "lastpass":
                if bw_entry:
                    bw_prefix = "LASTPASS_DEFAULT_PASSWORD='" + bw_entry.password + "' "
                    if bw_entry.totp_now:
                        bw_prefix += "LASTPASS_DEFAULT_OTP='" + bw_entry.totp_now + "' "
                elif lp_entry:
                    bw_prefix = "LASTPASS_DEFAULT_PASSWORD='" + lp_entry.password + "' "
                if "ndt_mfa_token" in profile_data:
                    bw_prefix += (
                        "LASTPASS_DEFAULT_OTP='"
                        + mfa_generate_code(profile_data["ndt_mfa_token"])
                        + "' "
                    )
                print(
                    bw_prefix
                    + "lastpass-aws-login --profile "
                    + profile
                    + " --no-prompt"
                )
        elif "AWS_SESSION_EXPIRATION_EPOC_" + safe_profile not in os.environ:
            print_profile_expiry(profile)
    elif profile_type == "ndt":
        if "AWS_SESSION_EXPIRATION_EPOC_" + safe_profile in os.environ:
            expiry = int(os.environ["AWS_SESSION_EXPIRATION_EPOC_" + safe_profile])
        if expiry < now:
            expiry = read_profile_expiry_epoc(profile)
        if expiry < now:
            if "AWS_SESSION_EXPIRATION_EPOC_" + safe_profile in os.environ:
                print("unset AWS_SESSION_EXPIRATION_EPOC_" + safe_profile + ";")
            profile_data = get_profile(profile)
            if "ndt_origin_profile" not in profile_data:
                return
            origin_profile = profile_data["ndt_origin_profile"]
            origin_profile_data = get_profile(origin_profile)
            if "azure_tenant_id" in origin_profile_data:
                origin_type = "azure"
            elif "adfs_login_url" in origin_profile_data:
                origin_type = "adfs"
            elif "lastpass_saml_id" in origin_profile_data:
                origin_type = "lastpass"
            else:
                origin_type = "iam"
            enable_profile(origin_type, origin_profile)

            command = ["ndt", "assume-role"]
            if "ndt_mfa_token" in profile_data:
                command.append("-t")
                command.append(profile_data["ndt_mfa_token"])
            if "ndt_default_duration_hours" in profile_data:
                command.append("-d")
                duration = _to_str(int(profile_data["ndt_default_duration_hours"]) * 60)
                command.append(duration)
            command.append("-p")
            command.append(profile)
            command.append(profile_data["ndt_role_arn"])
            print(" ".join(command))
        elif "AWS_SESSION_EXPIRATION_EPOC_" + safe_profile not in os.environ:
            print_profile_expiry(profile)
        _print_profile_switch(profile)
    elif profile_type == "azure-subscription":
        if not (
            "AZURE_SUBSCRIPTION" in os.environ
            and os.environ["AZURE_SUBSCRIPTION"] == orig_profile
        ):
            subscription_id = az_select_subscription(orig_profile)
            if subscription_id:
                print('AZURE_SUBSCRIPTION="' + orig_profile + '";')
                print('AZURE_SUBSCRIPTION_ID="' + subscription_id + '";')
                print("export AZURE_SUBSCRIPTION AZURE_SUBSCRIPTION_ID")


def _print_profile_switch(profile):
    unset = []
    for env in ["AWS_SESSION_TOKEN", "AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"]:
        if env in os.environ:
            unset.append(env)
    if unset:
        print("unset " + " ".join(unset) + ";")
    set_env = []
    if (
        "AWS_DEFAULT_PROFILE" not in os.environ
        or os.environ["AWS_DEFAULT_PROFILE"] != profile
    ):
        set_env.append("AWS_DEFAULT_PROFILE")
    if "AWS_PROFILE" not in os.environ or os.environ["AWS_PROFILE"] != profile:
        set_env.append("AWS_PROFILE")
    if set_env:
        for param in set_env:
            print(param + '="' + profile + '";')
        print("export " + " ".join(set_env) + ";")


def _epoc_secs(d):
    return int(
        (d - datetime.utcfromtimestamp(0).replace(tzinfo=tzutc())).total_seconds()
    )



import time
from threadlocal_aws.clients import route53


def upsert_record(dns_name, record_type, value, ttl=300, wait=True):
    zone_id, change = get_record_change(
        dns_name, record_type, value, hosted_zones(), ttl=ttl
    )
    request = route53().change_resource_record_sets(
        HostedZoneId=zone_id, ChangeBatch={"Changes": [change]}
    )["ChangeInfo"]
    if wait:
        status = route53().get_change(Id=request["Id"])["ChangeInfo"]["Status"]
        while not status == "INSYNC":
            print(
                "Waiting for request to sync - "
                + str(request["Id"])
                + "("
                + dns_name
                + ") status: "
                + status
            )
            time.sleep(2)
            status = route53().get_change(Id=request["Id"])["ChangeInfo"]["Status"]


def hosted_zones():
    pages = route53().get_paginator("list_hosted_zones")
    for page in pages.paginate():
        for hosted_zone in page.get("HostedZones", []):
            yield hosted_zone


def longest_matching_zone(alias, hosted_zones):
    ret = {"Name": ""}
    for zone in hosted_zones:
        if (alias + ".").endswith(zone["Name"]) and len(zone["Name"]) > len(
            ret["Name"]
        ):
            ret = zone
    return ret


def get_record_change(dns_name, record_type, value, hosted_zones, ttl=300):
    zone = longest_matching_zone(dns_name, hosted_zones)
    if zone:
        print(
            dns_name
            + " => "
            + "("
            + record_type
            + ") '"
            + value
            + "' in "
            + zone["Name"]
        )
        change = {
            "Action": "UPSERT",
            "ResourceRecordSet": {
                "Name": dns_name,
                "Type": record_type,
                "TTL": ttl,
                "ResourceRecords": [{"Value": value}],
            },
        }
        return zone["Id"], change



# Copyright 2016-2017 Nitor Creations Oy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from setuptools import setup
from n_utils import PATH_COMMANDS, CONSOLESCRIPTS

win_deps = []
if sys.version_info[0] == 2:
    python2_or_3_deps = [
        "pyotp==2.3",
        "Pygments==2.5.2",
        "importlib-metadata==2.1.1",
        "decorator==4.4.2",
        "PyYAML==5.2",
    ]
    python2_or_3_test_deps = ["pytest==4.6.11", "pytest-mock==1.13.0", "mock==3.0.5"]
elif sys.version_info[0] == 3:
    python2_or_3_deps = ["pyotp", "Pygments", "PyYAML>=5.2"]
    python2_or_3_test_deps = ["pytest-mock", "mock"]
    python2_or_3_test_deps.insert(0, "pytest")
if sys.platform.startswith("win"):
    win_deps = ["win-unicode-console", "wmi", "pypiwin32"]

with open("README.md") as f:
    long_description = f.read()

setup(
    name="nameless-deploy-tools",
    version="1.232",
    description="Tools for deploying to AWS via CloudFormation and Serverless framework that support a pull request based workflow",
    url="http://github.com/NitorCreations/nameless-deploy-tools",
    download_url="https://github.com/NitorCreations/nameless-deploy-tools/tarball/1.157",
    author="Pasi Niemi",
    author_email="pasi@nitor.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    packages=["n_utils"],
    include_package_data=True,
    scripts=PATH_COMMANDS,
    entry_points={
        "console_scripts": CONSOLESCRIPTS,
    },
    install_requires=[
        "future",
        "pyaml",
        "awscli",
        "requests",
        "termcolor",
        "ipaddr",
        "argcomplete",
        "nitor-vault>=0.41",
        "pyqrcode",
        "six",
        "python-dateutil",
        "pycryptodomex",
        "configparser",
        "scandir",
        "jmespath",
        "ec2-utils>=0.34",
        "cloudformation-utils==0.0.2",
        "pyOpenSSL>=19.1.0",
    ]
    + python2_or_3_deps
    + win_deps,
    tests_require=["pytest-cov", "coverage"] + python2_or_3_test_deps,
    zip_safe=False,
)



import os
import json
from collections import OrderedDict
from threadlocal_aws.clients import connect
from n_utils.aws_infra_util import (
    yaml_to_dict,
    yaml_save,
    json_load,
    json_save,
    json_save_small,
    load_parameters,
    import_scripts,
)
from n_utils.ndt_project import Project


def deploy_connect_contact_flows(component, contactflowname, dry_run=True):
    subcomponent = (
        Project().get_component(component).get_subcomponent("connect", contactflowname)
    )
    extra_parameters = load_parameters(component=component, connect=contactflowname)
    template = subcomponent.get_dir() + os.sep + "template.yaml"
    flow_defs = yaml_to_dict(template, extra_parameters=extra_parameters)
    if "connectInstanceId" not in flow_defs:
        raise Exception(
            "Missing connect instance id. Define 'connectInstanceId' in the root of your flows template."
        )

    instance_id = flow_defs["connectInstanceId"]
    flows = get_flows(instance_id)
    if "contactFlows" in flow_defs:
        for flow in flow_defs["contactFlows"]:
            import_scripts(flow, template, extra_parameters=extra_parameters)
            if flow["Name"] in flows:
                flow_id = flows[flow["Name"]]["Id"]
                flow_type = flow["Type"]
                flow_name = flow["Name"]
                flow_tags = flow["Tags"]
                del flow["Name"]
                del flow["Type"]
                del flow["Tags"]
                if "Description" in flow:
                    flow_description = flow["Description"]
                    del flow["Description"]
                flow_content = json_save_small(flow)
                print("Updating flow " + flow_name)
                if not dry_run:
                    try:
                        connect().update_contact_flow_content(
                            InstanceId=instance_id,
                            ContactFlowId=flow_id,
                            Content=flow_content,
                        )
                        connect().update_contact_flow_name(
                            InstanceId=instance_id,
                            ContactFlowId=flow_id,
                            Description=flow_description,
                            Name=flow_name,
                        )
                    except:
                        print("Failed to update flow " + flow_name)
                        print(yaml_save(flow))
            else:
                flow_type = flow["Type"]
                flow_name = flow["Name"]
                flow_tags = flow["Tags"]
                flow_description = None
                del flow["Name"]
                del flow["Type"]
                del flow["Tags"]
                if "Description" in flow:
                    flow_description = flow["Description"]
                    del flow["Description"]
                flow_content = json_save_small(flow)
                print("Creating flow " + flow_name)
                if not dry_run:
                    try:
                        connect().create_contact_flow(
                            InstanceId=instance_id,
                            Name=flow_name,
                            Type=flow_type,
                            Description=flow_description,
                            Content=flow_content,
                            Tags=flow_tags,
                        )
                    except:
                        print("Failed to create flow " + flow_name)
                        print(yaml_save(flow))


def export_connect_contact_flow(instance_id, flowname):
    paginator = connect().get_paginator("list_contact_flows")
    for page in paginator.paginate(InstanceId=instance_id):
        for flow in page["ContactFlowSummaryList"]:
            if flow["Name"] == flowname:
                content = OrderedDict()
                contact_flow = connect().describe_contact_flow(
                    InstanceId=instance_id, ContactFlowId=flow["Id"]
                )["ContactFlow"]
                content["Name"] = contact_flow["Name"]
                content["Type"] = contact_flow["Type"]
                content["Tags"] = contact_flow["Tags"]
                if "Description" in contact_flow:
                    content["Description"] = contact_flow["Description"]
                content.update(json_load(contact_flow["Content"]))
                return yaml_save(content)


def get_instance_ids():
    ret = []
    paginator = connect().get_paginator("list_instances")
    for page in paginator.paginate():
        for instance in page["InstanceSummaryList"]:
            ret.append(instance["Id"])
    return ret


def get_instance_aliases():
    ret = []
    paginator = connect().get_paginator("list_instances")
    for page in paginator.paginate():
        for instance in page["InstanceSummaryList"]:
            ret.append(instance["InstanceAlias"])
    return ret


def alias_to_id(alias):
    paginator = connect().get_paginator("list_instances")
    for page in paginator.paginate():
        for instance in page["InstanceSummaryList"]:
            if instance["InstanceAlias"] == alias:
                return instance["Id"]
    return None


def get_flows(instance_id):
    existing_flows = OrderedDict()
    paginator = connect().get_paginator("list_contact_flows")
    for page in paginator.paginate(InstanceId=instance_id):
        for flow in page["ContactFlowSummaryList"]:
            existing_flows[flow["Name"]] = flow
    return existing_flows
